#!/usr/bin/env python3

from ChildProject.projects import ChildProject
from ChildProject.annotations import AnnotationManager
from ChildProject.metrics import segments_to_annotation

import argparse
import datalad.api
from os.path import join as opj
from os.path import basename, exists
import pandas as pd
import pickle
from pyannote.core import Annotation, Segment, Timeline

import stan
import numpy as np
#from scipy.stats import beta

parser = argparse.ArgumentParser(description = 'main model described throughout the notes.')
parser.add_argument('--group', default = 'child', choices = ['corpus', 'child'])
parser.add_argument('--chains', default = 4, type = int)
parser.add_argument('--samples', default = 2000, type = int)
parser.add_argument('--validation', default = 0, type = float)
parser.add_argument('--time-scale', default = 15, type = int)
parser.add_argument('--output', default = 'model1')
args = parser.parse_args()

def extrude(self, removed, mode: str = 'intersection'):
    if isinstance(removed, Segment):
        removed = Timeline([removed])

    truncating_support = removed.gaps(support=self.extent())
    # loose for truncate means strict for crop and vice-versa
    if mode == "loose":
        mode = "strict"
    elif mode == "strict":
        mode = "loose"
    
    return self.crop(truncating_support, mode=mode)

def compute_counts(parameters):
    annotator = parameters['annotator']
    corpus = parameters['corpus']

    speakers = ['CHI', 'FEM', 'MAL', 'OCH']

    project = ChildProject(parameters['path'])
    am = AnnotationManager(project)
    am.read()

    intersection = AnnotationManager.intersection(
        am.annotations, ['vtc', annotator]
    )
    intersection['onset'] = intersection.apply(lambda r: np.arange(r['range_onset'], r['range_offset'], args.time_scale*1000), axis = 1)
    intersection = intersection.explode('onset')
    intersection['range_onset'] = intersection['onset']
    intersection['range_offset'] = (intersection['range_onset']+args.time_scale*1000).clip(upper = intersection['range_offset'])

    intersection['path'] = intersection.apply(
        lambda r: opj(project.path, 'annotations', r['set'], 'converted', r['annotation_filename']),
        axis = 1
    )
    intersection = intersection.merge(project.recordings[['recording_filename', 'child_id']], how = 'left')
    intersection['child'] = corpus + '_' + intersection['child_id'].astype(str)

    datalad.api.get(list(intersection['path'].unique()))

    segments = am.get_collapsed_segments(intersection)
    segments = segments[segments['speaker_type'].isin(speakers)]

    vtc = {
        speaker: segments_to_annotation(segments[(segments['set'] == 'vtc') & (segments['speaker_type'] == speaker)], 'speaker_type').get_timeline()
        for speaker in speakers
    }

    truth = {
        speaker: segments_to_annotation(segments[(segments['set'] == annotator) & (segments['speaker_type'] == speaker)], 'speaker_type').get_timeline()
        for speaker in speakers
    }

    print(parameters['corpus'], annotator)

    vtc = {key: len(vtc[key]) for key in vtc}

    segments['set'] = segments['set'].replace({annotator: 'truth'})
    segments = segments[segments['segment_offset'] > segments['segment_onset'] + 100]
    segments['corpus'] = corpus
    
    return (
        segments
        .groupby(['set', 'recording_filename', 'range_onset', 'speaker_type'])
        .agg(
            count = ('segment_onset', 'count'),
            child = ('child', lambda x: x.iloc[0])
        )
        .reset_index()
        .pivot(index = ['recording_filename', 'child', 'range_onset'], columns = ['set', 'speaker_type'], values = ['count'])
        .assign(
            corpus = parameters['corpus']
        )
    )

annotators = pd.read_csv('input/annotators.csv')[:-8]
annotators = annotators[~annotators['annotator'].str.startswith('eaf_2021/')]
annotators['path'] = annotators['corpus'].apply(lambda c: opj('input', c))
counts = pd.concat([compute_counts(annotator) for annotator in annotators.to_dict(orient = 'records')])
counts = counts.fillna(0)
counts.to_csv('counts.csv')

truth = np.transpose([counts['count']['truth'][speaker].values for speaker in ['CHI', 'OCH', 'FEM', 'MAL']]).astype(int)
vtc = np.transpose([counts['count']['vtc'][speaker].values for speaker in ['CHI', 'OCH', 'FEM', 'MAL']]).astype(int)

counts.reset_index(inplace = True)

rates = pd.read_csv('output/speech_dist.csv')

data = {
    'n_clips': truth.shape[0],
    'n_classes': truth.shape[1],
    'n_groups': counts[args.group].nunique(),
    'n_validation': max(1, int(truth.shape[0]*args.validation)),
    'n_sim': 100,
    'group': 1+counts[args.group].astype('category').cat.codes.values,
    'truth': truth.astype(int),
    'vtc': vtc.astype(int),
    'rates_alphas': rates['alpha'].values,
    'rates_betas': rates['beta'].values
}

print(f"clips: {data['n_clips']}")
print(f"groups: {data['n_groups']}")
print("true vocs: {}".format(np.sum(data['truth'])))
print("vtc vocs: {}".format(np.sum(data['vtc'])))

with open(f'data_{args.output}.pickle', 'wb') as fp:
    pickle.dump(data, fp, pickle.HIGHEST_PROTOCOL)

stan_code = """
data {
  int<lower=1> n_clips;   // number of clips
  int<lower=1> n_groups; // number of groups
  int<lower=1> n_classes; // number of classes
  int group[n_clips];
  int vtc[n_clips,n_classes];
  int truth[n_clips,n_classes];

  int<lower=1> n_validation;
  int<lower=1> n_sim;

  real<lower=0> rates_alphas[n_classes];
  real<lower=0> rates_betas[n_classes];
}

parameters {
  matrix<lower=0,upper=1>[n_classes,n_classes] mus;
  matrix<lower=1>[n_classes,n_classes] etas;
  //matrix<lower=0>[n_clips,n_classes] _lambdas;
  vector<lower=0>[n_classes] _lambdas;
  matrix<lower=0,upper=1>[n_classes,n_classes] group_confusion[n_groups];
}

transformed parameters {
  vector<lower=0>[n_classes] _p;
  matrix<lower=0>[n_classes,n_classes] alphas;
  matrix<lower=0>[n_classes,n_classes] betas;

  matrix[n_clips, n_classes] _log_lik;
  _log_lik = rep_matrix(0, n_clips, n_classes);

  alphas = mus * etas;
  betas = (1-mus) * etas;
  
  for (i in 1:n_classes) {
      _p[i] = 1/(1+1/(_lambdas[i]));
  }

    for (k in 1:n_clips) {
        vector [4] bp;
        for (i in 1:n_classes) {
            int n = 1;
            vector[200] log_contrib_comb;
            log_contrib_comb = rep_vector(0, 200);
            for (chi in 0:truth[k,1]) {
                bp[1] = binomial_lpmf(chi | truth[k,1], group_confusion[group[k],1,i]);
                for (och in 0:truth[k,2]) {
                    bp[2] = binomial_lpmf(och | truth[k,2], group_confusion[group[k],2,i]);
                    for (fem in 0:truth[k,3]) {
                        bp[3] = binomial_lpmf(fem | truth[k,3], group_confusion[group[k],3,i]);
                        for (mal in 0:truth[k, 4]) {
                            bp[4] = binomial_lpmf(mal | truth[k,4], group_confusion[group[k],4,i]);
                            int delta = vtc[k,i] - (mal+fem+och+chi);
                            if (delta >= 0) {
                                log_contrib_comb[n] += sum(bp);
                                log_contrib_comb[n] += log(1-_p[i]) + delta * log(_p[i]);
                                n = n+1;
                            }
                        }
                    }
                }
            }
            _log_lik[k,i] = log_sum_exp(log_contrib_comb[1:n]);
        }
    }
}

model {
    for (k in 1:n_clips) {
        target += log_sum_exp(_log_lik[k,:]);
    }

    //for (k in 1:n_clips) {
    //    for (i in 1:n_classes) {
    //        _lambdas[k,i] ~ exponential(1);
    //    }
    //}
    _lambdas ~ exponential(1);

    for (i in 1:n_classes) {
        for (j in 1:n_classes) {
            mus[i,j] ~ beta(1,1);
            etas[i,j] ~ pareto(1,1.5);
        }
    }

    for (c in 1:n_groups) {
        for (i in 1:n_classes) {
            for (j in 1:n_classes) {
                group_confusion[c,i,j] ~ beta(alphas[i,j], betas[i,j]);
            }
        }
    }
}

generated quantities {
    int pred[n_clips,n_classes,n_classes];
    matrix[n_classes,n_classes] probs[n_groups];

    int sim_truth[n_sim,n_classes];
    int sim_vtc[n_sim,n_classes];
    vector[n_classes] lambdas;
    real chi_adu_coef;

    if (uniform_rng(0,1) > 0.2) {
        chi_adu_coef = uniform_rng(0,1);
    }
    else {
        chi_adu_coef = 0;
    }

    for (c in 1:n_groups) {
        for (i in 1:n_classes) {
            for (j in 1:n_classes) {
                probs[c,i,j] = beta_rng(alphas[i,j], betas[i,j]);
            }
        }
    }

    for (k in 1:n_clips) {
        for (i in 1:n_classes) {
            for (j in 1:n_classes) {
                if (k >= n_validation) {
                    pred[k,i,j] = binomial_rng(truth[k,j], group_confusion[group[k],i,j]);
                }
                else {
                    pred[k,i,j] = binomial_rng(truth[k,j], probs[group[k],j,i]);
                }
            }
        }
    }

    real lambda;
    for (k in 1:n_sim) {
        for (i in 2:n_classes) {
            lambda = gamma_rng(rates_alphas[i], rates_betas[i]);
            sim_truth[k,i] = poisson_rng(lambda);
        }
        lambda = gamma_rng(rates_alphas[1], rates_betas[1]);
        sim_truth[k,1] = poisson_rng(lambda + chi_adu_coef*(sim_truth[k,3]+sim_truth[k,4]));
    }

    for (k in 1:n_sim) {
        for (i in 1:n_classes) {
            sim_vtc[k,i] = poisson_rng(exponential_rng(1/_lambdas[i]));
            for (j in 1:n_classes) {
                real p = beta_rng(alphas[j,i], betas[j,i]);
                sim_vtc[k,i] += binomial_rng(sim_truth[k,j], p);
            }
        }
    }
}
"""

posterior = stan.build(stan_code, data = data)
fit = posterior.sample(num_chains = args.chains, num_samples = args.samples)
df = fit.to_frame()
df.to_parquet(f'fit_{args.output}.parquet')
