#!/usr/bin/env python3

import pandas as pd
import pickle
import numpy as np
import pyarrow.parquet as pq

import argparse
import matplotlib
import matplotlib.pyplot as plt
matplotlib.use("pgf")
matplotlib.rcParams.update({
    "pgf.texsystem": "xelatex",
    'font.family': 'serif',
    "font.serif" : "Times New Roman",
    'text.usetex': True,
    'pgf.rcfonts': False,
})

from sklearn.linear_model import LinearRegression

from tqdm import tqdm

def set_size(width, fraction=1, ratio = None):
    fig_width_pt = width * fraction
    inches_per_pt = 1 / 72.27
    if ratio is None:
        ratio = (5 ** 0.5 - 1) / 2
    fig_width_in = fig_width_pt * inches_per_pt
    fig_height_in = fig_width_in * ratio
    return fig_width_in, fig_height_in


df = pd.read_csv("input/vanuatu-paper/correlations.csv").set_index("speaker")
df = df.to_dict(orient="index")
print(df)

parser = argparse.ArgumentParser(description = 'plot_sim_null')
parser.add_argument('--data', nargs='+')
parser.add_argument('--fit', nargs='+')
parser.add_argument('--output')
args = parser.parse_args()

data = []

for i, _data in enumerate(args.data):
    with open(_data, 'rb') as fp:
        data.append(pickle.load(fp))

n_samples = 0

for i, _fit in enumerate(args.fit):
    table = pq.read_table(_fit, columns=[])
    n_samples += table.num_rows


speakers = ['CHI', 'OCH', 'FEM', 'MAL']

#n_simulations = fit['n_sim'].iloc[0]
n_sim = data[0]['n_sim']
n_sim = 38

sim_beta_fem = np.zeros(n_samples)
sim_beta_mal = np.zeros(n_samples)
sim_beta_och = np.zeros(n_samples)
sim_r_fem = np.zeros(n_samples)
sim_r_mal = np.zeros(n_samples)
sim_r_och = np.zeros(n_samples)
true = np.zeros(n_samples)


pbar = tqdm(total=n_samples)

seek = 0
for i, _fit in enumerate(args.fit):
    fit = pd.read_parquet(_fit)
    fit = fit[fit['chi_adu_coef'] < 0.0001]

    for j in range(len(fit)):
        f = fit.iloc[j]

        chi_vtc = np.array([f[f'sim_vtc.{k+1}.1'] for k in range(n_sim)])
        fem_vtc = np.array([f[f'sim_vtc.{k+1}.3'] for k in range(n_sim)])
        mal_vtc = np.array([f[f'sim_vtc.{k+1}.4'] for k in range(n_sim)])
        och_vtc = np.array([f[f'sim_vtc.{k+1}.2'] for k in range(n_sim)])

        regr = LinearRegression()
        regr.fit(fem_vtc.reshape(-1, 1), chi_vtc)
        sim_beta_fem[seek+j] = regr.coef_[0]
        sim_r_fem[seek+j] = np.corrcoef(fem_vtc, chi_vtc)[0,1]

        regr = LinearRegression()
        regr.fit(mal_vtc.reshape(-1, 1), chi_vtc)
        sim_beta_mal[seek+j] = regr.coef_[0]
        sim_r_mal[seek+j] = np.corrcoef(mal_vtc, chi_vtc)[0,1]

        regr = LinearRegression()
        regr.fit(och_vtc.reshape(-1, 1), chi_vtc)
        sim_beta_och[seek+j] = regr.coef_[0]
        sim_r_och[seek+j] = np.corrcoef(och_vtc, chi_vtc)[0,1]

        pbar.update(1)   

    seek += len(fit)

pbar.close()

fig, ax = plt.subplots(1,2,figsize=set_size(450,1,0.5))
ax[0].hist(sim_beta_fem, histtype = 'step', bins = 20, density = True, label = '$\\mathrm{CHI} = K \\mathrm{FEM} + I$')
ax[0].hist(sim_beta_mal, histtype = 'step', bins = 20, density = True, label = '$\\mathrm{CHI} = K \\mathrm{MAL} + I$')
ax[0].hist(sim_beta_och, histtype = 'step', bins = 20, density = True, label = '$\\mathrm{CHI} = K \\mathrm{OCH} + I$')
ax[0].set_xlabel('$\\hat{K}$')
ax[0].set_ylabel('$p(\\hat{K}|K=0)$')
ax[0].legend(loc = 'upper left', bbox_to_anchor=[0, 1.25], frameon=False)
ax[0].set_xlim(-1,1)

ax[1].hist(sim_r_fem, histtype = 'step', bins = 20, density = True, label = '$R(\\mathrm{CHI},\\mathrm{FEM})$')
ax[1].hist(sim_r_mal, histtype = 'step', bins = 20, density = True, label = '$R(\\mathrm{CHI},\\mathrm{MAL})$')
ax[1].hist(sim_r_och, histtype = 'step', bins = 20, density = True, label = '$R(\\mathrm{CHI},\\mathrm{OCH})$')
ax[1].set_xlabel('$R$')
ax[1].set_ylabel('$p(R|K=0)$')
ax[1].legend(loc = 'upper left', bbox_to_anchor=[0, 1.25], frameon=False)
ax[1].set_xlim(-1,1)

fig.savefig(args.output, bbox_inches = 'tight')

df = pd.read_csv("input/vanuatu-paper/correlations.csv").set_index("speaker")
df = df.to_dict(orient="index")

from scipy import stats


df["fem"]["p"] = 100-stats.percentileofscore(sim_r_fem, df["fem"]["R"])
df["mal"]["p"] = 100-stats.percentileofscore(sim_r_mal, df["mal"]["R"])
df["och"]["p"] = 100-stats.percentileofscore(sim_r_och, df["och"]["R"])

print(len(sim_r_och[sim_r_och > df["och"]["R"]]))

print(df)

df = pd.DataFrame(df)
print(df)

df.to_csv("vanuatu_probs.csv")
